#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals, division
import sys
from datetime import datetime
from time import strftime

import six
import windbreads.utils as wdu
import wx
import wx.richtext as rt

RTC_ALIGNS = dict(
    default=wx.TEXT_ALIGNMENT_DEFAULT,
    left=wx.TEXT_ALIGNMENT_LEFT,
    centre=wx.TEXT_ALIGNMENT_CENTRE,
    center=wx.TEXT_ALIGNMENT_CENTER,
    right=wx.TEXT_ALIGNMENT_RIGHT,
)
CHINESE_FONTS = (
    "微软雅黑",
    "雅黑",
    "microsoft yahei",
    "黑体",
    "宋体",
    "新宋体",
    "仿宋",
    "楷体",
    "simhei",
    "heiti",
    "simsun",
    "nsimsun",
    "fangsong",
    "kaiti",
    "mingliu",
    "pmingliu",
)
FS_ENCODING = sys.getfilesystemencoding()


def get_text_width(text, wgt):
    font = wgt.GetFont()
    dc = wx.WindowDC(wgt)
    dc.SetFont(font)
    return dc.GetTextExtent(text)


def get_adjust_size(size=(-1, -1), **kwargs):
    if size == (-1, -1):
        return size

    ratio_size = kwargs.get("ratio_size", (1600, 900))
    w, h = size
    rw, rh = ratio_size
    dw, dh = wx.GetDisplaySize()
    if rw == dw and rh == dh:
        return size

    bw, bh = -1, -1
    if w != -1:
        bw = int((w * dw) / rw)

    if h != -1:
        bh = int((h * dh) / rh)

    return (bw, bh)


def require_int(evt, min_value=1, max_value=-1):
    wgt = evt.GetEventObject()
    text = wgt.GetValue().strip()
    if text.isdigit():
        wgt.ChangeValue(text)
        wgt.SetInsertionPoint(wgt.GetInsertionPoint())


def auto_get_font(obj=None, **kwargs):
    if "font" in kwargs:
        return kwargs["font"]

    try:
        return obj.auto_font()
    except Exception:
        pass

    return None


def get_chinese_fonts():
    e = wx.FontEnumerator()
    e.EnumerateFacenames()
    fonts = []
    for name in (f for f in e.GetFacenames() if not f.startswith("@")):
        if name.lower().startswith(CHINESE_FONTS):
            fonts.append(name)

    return fonts


def get_best_chinese_font(fonts=[]):
    for cf in CHINESE_FONTS:
        for font in fonts:
            if font.lower().startswith(cf):
                return font

    return None


def pydate2wxdate(date):
    tt = date.timetuple()
    dmy = (tt[2], tt[1] - 1, tt[0])
    return wx.DateTimeFromDMY(*dmy)


def wxdate2pydate(date):
    if date.IsValid():
        return datetime.date(*(map(int, date.FormatISODate().split("-"))))

    return None


def cat_echo_text(**kwargs):
    args = kwargs.get("args")
    kargs = kwargs.get("kargs")
    t = kwargs.get("t")
    text = kwargs.get("text", "")
    if text:
        if kargs:
            return wdu.ttt(text, t).format(**kargs)

        if args:
            if not isinstance(args, (tuple, list)):
                args = (args,)

            return wdu.ttt(text, t).format(*args)

        return wdu.ttt(text, t)

    return text


def write_echo_text(**kwargs):
    ts_text = kwargs.get("ts_text")
    if kwargs.get("tff"):  # t for file
        text = cat_echo_text(**kwargs)
    else:
        kargs = kwargs.copy()
        kargs.update(t=None)
        text = cat_echo_text(**kargs)

    encoding = kwargs.get("encoding", "utf-8")
    if six.PY2 and isinstance(text, six.text_type):
        text = text.encode(encoding)
    elif six.PY3:
        text = bytes(text, encoding)
        if ts_text:
            ts_text = bytes(ts_text, encoding)

    nl = kwargs.get("nl", True)
    log_mode = kwargs.get("log_mode", "a" if six.PY2 else "ab")

    log_file = kwargs.get("log_file")
    log_files = kwargs.get("log_files", [])
    if log_file:
        log_files = [log_file]

    for log_file in log_files:
        with open(log_file, log_mode) as f:
            if ts_text:
                f.write(ts_text)

            f.write(text)
            if nl:
                f.write(wdu.NEW_LINE)


def echo_text(
    rtc,
    text="",
    fg=None,
    bg=None,
    ts=True,
    nl=True,
    italic=False,
    align=None,
    underline=False,
    bold=False,
    ts_style=False,
    font=None,
    size=None,
    clear=False,
    keep_date=True,
    **kwargs
):
    if ts:
        now = datetime.now()
        if keep_date:
            ts_text = "[{}] ".format(now)
        else:
            ts_text = "[{}] ".format(str(now).split(" ", 1)[-1])

    else:
        ts_text = ""

    if text and isinstance(text, six.string_types):
        if not isinstance(text, six.text_type):
            utext = text.decode(wdu.detect_encoding(text)["encoding"])
        else:
            utext = text

    else:
        utext = "{}".format(text)

    write_echo_text(ts_text=ts_text, text=utext, nl=nl, **kwargs)
    if kwargs.get("no_echo", False):
        if clear:
            rtc.Clear()

        return

    rtc.SetInsertionPointEnd()
    rta = rt.RichTextAttr()
    rta.SetAlignment(RTC_ALIGNS["default"])
    rta.SetTextColour("black")
    rta.SetBackgroundColour("white")
    rta.SetFontStyle(wx.FONTSTYLE_NORMAL)
    rta.SetFontWeight(wx.FONTWEIGHT_NORMAL)
    rta.SetFontUnderlined(False)
    rtc.SetDefaultStyle(rta)
    if ts_text and not ts_style:
        rtc.WriteText(ts_text)

    align = RTC_ALIGNS.get(align)
    if align:
        rta.SetAlignment(align)

    if fg:
        rta.SetTextColour(fg)

    if bg:
        rta.SetBackgroundColour(bg)

    if font:
        rta.SetFontFaceName(font)

    if size:
        rta.SetFontSize(size)

    if bold is True:
        rta.SetFontWeight(wx.FONTWEIGHT_BOLD)
    elif bold is False:
        rta.SetFontWeight(wx.FONTWEIGHT_NORMAL)

    if italic is True:
        rta.SetFontStyle(wx.FONTSTYLE_ITALIC)
    elif italic is False:
        rta.SetFontStyle(wx.FONTSTYLE_NORMAL)

    if underline is not None:
        rta.SetFontUnderlined(underline)

    rtc.BeginStyle(rta)

    if ts_text and ts_style:
        rtc.WriteText(ts_text)

    rtc.WriteText(cat_echo_text(text=utext, **kwargs))
    rtc.EndStyle()
    if nl:
        rtc.Newline()

    rtc.ShowPosition(rtc.GetLastPosition())
    if clear:
        rtc.Clear()


def on_hide(self, evt=None):
    self.Hide()


def on_echoing(self, **kwargs):
    """Default method for echoing text."""
    if self.is_echoing or not self.echo_lines:
        return

    self.is_echoing = True
    i = 0
    batch = kwargs.get("batch", 100)
    while i < batch and self.echo_lines:
        line = self.echo_lines.pop(0)
        self.echo_text(line[0], **line[1])
        i += 1

    self.is_echoing = False


def start_timer(timer, miliseconds=1000, one_shot=False):
    if timer.IsRunning():
        return

    timer.Start(int(miliseconds), one_shot)


def stop_timer(timer, delete=False):
    if timer.IsRunning():
        timer.Stop()

    if delete:
        del timer


def stop_timers(timers=[], delete=False):
    [stop_timer(timer, delete) for timer in timers]


def update_clock_statusbar(sbar, ts_fmt="%Y/%m/%d %H:%M", idx=2):
    text = strftime(ts_fmt)
    try:
        text = text.decode(FS_ENCODING)
    except AttributeError:
        pass

    set_status_text(sbar, text.encode("utf-8"), idx)


def set_status_text(sbar, text, idx, t=None):
    sbar.SetStatusText(t(text) if t else text, idx)


def on_popup_lang(self, evt):
    if not hasattr(self, "english_id"):
        self.english_id = wx.NewIdRef()
        self.chinese_id = wx.NewIdRef()

        self.Bind(wx.EVT_MENU, self.popup_lang, id=self.english_id)
        self.Bind(wx.EVT_MENU, self.popup_lang, id=self.chinese_id)

    menu = wx.Menu()
    menu.Append(self.english_id, "English", "", wx.ITEM_RADIO)
    menu.Append(self.chinese_id, "Chinese - 简体中文", "", wx.ITEM_RADIO)
    if self.lang == "zh":
        menu.Check(self.chinese_id, True)
    else:
        menu.Check(self.english_id, True)

    self.PopupMenu(menu)
    menu.Destroy()


def popup_lang(self, evt):
    lang = "zh" if evt.GetId() == self.chinese_id else "en"
    if lang != self.lang:
        self.lang = lang
        self.update_t()
        self.update_ui_lang()
        self.save_lang(lang)

    evt.Skip()


def update_ui_lang(self, refresh=True):
    import wxbreads.widgets as wxw

    if hasattr(self, "lang_wgts"):
        for lwgt in self.lang_wgts:
            tooltip = ""
            if len(lwgt) == 2:
                wgt, label = lwgt
            else:
                wgt, label, tooltip = lwgt

            wxw.set_tooltip(wgt, tooltip, self.t)
            wgt.SetLabel(self.tt(label))

    if refresh:
        if hasattr(self, "panel"):
            self.panel.Layout()

        self.Refresh()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import six
import wx
import windbreads.utils as wdu

try:
    TaskBarIcon = wx.TaskBarIcon
    IconFromBitmap = wx.IconFromBitmap
    EVT_TASKBAR_LEFT_DOWN = wx.EVT_TASKBAR_LEFT_DOWN
except AttributeError:
    import wx.adv

    TaskBarIcon = wx.adv.TaskBarIcon
    IconFromBitmap = wx.Icon
    EVT_TASKBAR_LEFT_DOWN = wx.adv.EVT_TASKBAR_LEFT_DOWN


class TrayIcon(TaskBarIcon):
    restore_id = wx.NewIdRef()
    quit_id = wx.NewIdRef()

    def __init__(self, frame, icon=None, text="TrayIcon", **kwargs):
        self.t = kwargs.get("t")
        TaskBarIcon.__init__(self)
        self.frame = frame
        if icon:
            if not isinstance(icon, six.string_types):
                icon = self.make_icon(icon)

            self.SetIcon(icon, wdu.ttt(text, self.t))

        self.Bind(EVT_TASKBAR_LEFT_DOWN, self.on_restore)
        self.Bind(wx.EVT_MENU, self.on_restore, id=self.restore_id)
        self.Bind(wx.EVT_MENU, self.on_quit, id=self.quit_id)

    def CreatePopupMenu(self):
        menu = wx.Menu()
        menu.Append(self.restore_id, wdu.ttt("Restore", self.t))
        menu.AppendSeparator()
        menu.Append(self.quit_id, wdu.ttt("Quit", self.t))
        return menu

    def make_icon(self, img):
        if "wxMSW" in wx.PlatformInfo:
            img = img.Scale(16, 16)
        elif "wxGTK" in wx.PlatformInfo:
            img = img.Scale(22, 22)

        # wxMac can be any size upto 128x128, so leave the source img alone....
        icon = IconFromBitmap(img.ConvertToBitmap())
        return icon

    def on_restore(self, event):
        if self.frame.IsIconized():
            self.frame.Iconize(False)

        if not self.frame.IsShown():
            self.frame.Show(True)

        self.frame.Raise()

    def on_quit(self, event):
        self.frame.on_quit(event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import ldap3
import windbreads.utils as wdu
import wx
from ldap3tool import LDAPTool

import wxbreads.widgets as wxw
from wxbreads.base import BaseDialog


def ldap_login(server, base_dn, login_name, password, close=True, **kwargs):
    ldap = LDAPTool(server=server, base_dn=base_dn)
    try:
        use_ssl = kwargs.pop("use_ssl", False)
        kw = kwargs.copy()
        if use_ssl:
            kw.update(server=ldap.open_server(use_ssl=use_ssl))

        ldap.connect(login_name, password, **kw)
    except ldap3.core.exceptions.LDAPSocketOpenError as ex:
        return 100, ex
    except ldap3.core.exceptions.LDAPBindError:
        return 200, None
    except Exception as ex:
        return 300, ex
    else:
        if close:
            ldap.close()
        else:
            return 0, ldap

    return 0, None


class LoginWindow(BaseDialog):
    app_title = "User Login"

    def __init__(self, **kwargs):
        self.enable_cancel = kwargs.get("enable_cancel", True)
        self.ldap_kwargs = kwargs.pop("ldap_kwargs", {})
        self.need_busy = kwargs.pop("need_busy", False)
        self.allowed_names = kwargs.pop("allowed_names", None)

        super(LoginWindow, self).__init__(**kwargs)
        self.panel = wx.Panel(self)
        self.parent = parent = kwargs.get("parent")
        root_user = kwargs.get("root_user")
        root_pass = kwargs.get("root_pass")
        last_user = kwargs.get("last_user")
        self.domain = kwargs.get("domain", "")
        if parent:
            if not root_user:
                root_user = getattr(parent, "root_user", "root")

            if not root_pass:
                root_pass = getattr(parent, "root_pass", "")

            if not last_user:
                last_user = getattr(parent, "login_user")

        self.root_user = root_user or "root"
        self.root_pass = root_pass or ""
        self.last_user = last_user or ""
        self.pwd = kwargs.get("password", "")
        self.current_user = None
        self.ldap_obj = None

        self.server = kwargs.get("server", "ldap-server")
        self.base_dn = kwargs.get("base_dn", "dc=corp,dc=company,dc=org")
        self.destroy = kwargs.get("destroy", True)
        self.can_exit = kwargs.get("can_exit", True)
        self.is_login = False

        self.setup_ui()
        self.Bind(wx.EVT_BUTTON, self.on_login, id=wx.ID_OK)
        self.Bind(wx.EVT_BUTTON, self.on_quit, id=wx.ID_CANCEL)
        if kwargs.get("show_win", True):
            self.show()

    def setup_ui(self):
        sizer = wx.BoxSizer(wx.VERTICAL)
        size = (120, -1)
        label_style = wx.ALIGN_RIGHT
        kwargs = dict(
            fsize=size, ssize=(200, -1), fstyle=label_style, t=self.t
        )

        # Name field
        _, self.name_tc = wxw.add_text_row(
            self.panel,
            sizer,
            label="Login Name",
            value=self.last_user,
            **kwargs
        )
        wxw.focus_on(self.name_tc)

        # Password
        _, self.pwd_tc = wxw.add_text_row(
            self.panel,
            sizer,
            label="Password",
            value=self.pwd,
            sstyle=wx.TE_PASSWORD,
            **kwargs
        )

        ok_btn, cancel_btn = wxw.add_ok_buttons(
            self.panel, sizer, size=(100, 30), ok_text="Login", t=self.t
        )

        self.ok_btn = ok_btn
        self.cancel_btn = cancel_btn
        cancel_btn.Enable(self.enable_cancel)

        self.panel.SetSizer(sizer)
        self.panel.Layout()
        sizer.Fit(self)

    def high_light(self, wgt, focus=True):
        wgt.Clear()
        wgt.SetBackgroundColour(wxw.HIGHLIGHT_RED)
        if focus:
            wgt.SetFocus()

    def get_field_values(self):
        self.login_name = self.name_tc.GetValue().strip().lower()
        self.password = self.pwd_tc.GetValue()

    def on_login(self, event):
        self.ok_btn.Enable(False)
        self.cancel_btn.Enable(False)
        self.is_login = True
        self.current_user = None
        self.ldap_obj = None
        self.start_delay_work(self.after_submit, self.do_submit)

    def after_submit(self, delay_result):
        try:
            delay_result.get()
        except Exception as e:
            self.popup("Error", e, "e")

        self.ok_btn.Enable(True)
        self.cancel_btn.Enable(True)
        self.is_login = False
        if self.current_user:
            if self.parent and hasattr(self.parent, "login_user"):
                self.parent.login_user = self.current_user
                self.parent.ldap_obj = self.ldap_obj

            self.on_quit()

    def do_submit(self):
        self.get_field_values()
        if not self.login_name:
            self.high_light(self.name_tc)
            self.Refresh()
            return

        self.name_tc.SetBackgroundColour(wx.NullColour)
        if not self.password:
            self.name_tc.ClearBackground()
            self.high_light(self.pwd_tc)
            self.Refresh()
            return

        if self.parent:
            root_user = getattr(self.parent, "root_user", "root")
            root_pass = getattr(self.parent, "root_pass", "")
        else:
            root_user = self.root_user
            root_pass = self.root_pass

        if self.login_name == root_user:
            if self.password == root_pass:
                self.current_user = self.login_name
                self.ldap_obj = None
                return

            self.high_light(self.pwd_tc)
            self.Refresh()
            return

        if "\\" in self.login_name:
            domain, username = self.login_name.split("\\", 1)
        else:
            domain, username = self.domain, self.login_name

        if self.allowed_names and username not in self.allowed_names:
            self.popup("Error", "This user is not allowed to login", "e")
            return

        busy = None
        if self.need_busy:
            busy = self.show_busy("connecting to server...")

        ec, msg = ldap_login(
            self.server,
            self.base_dn,
            "{}\\{}".format(domain, username) if domain else username,
            self.password,
            **self.ldap_kwargs
        )
        if self.need_busy:
            del busy

        if ec in (100, 300):
            self.popup("Error", msg, "e")
            self.Refresh()
            return
        elif ec == 200:
            self.popup(
                "Authentication Error", "Username/password not match", "e"
            )
            self.high_light(self.pwd_tc)
            self.Refresh()
            return

        if self.parent and hasattr(self.parent, "extra_login_check"):
            if not self.parent.extra_login_check(username, msg):
                return

        self.current_user = username
        self.ldap_obj = msg

    def on_quit(self, event=None):
        if self.can_exit and (not self.is_login):
            if self.destroy:
                self.Destroy()
            else:
                self.Hide()


def test_run():
    app = wx.App()
    LoginWindow(last_user="domain\\username", password="password")
    app.MainLoop()


def test_i18n_run():
    from functools import partial

    import windbreads.common_i18n as wdi18n

    zh = wdi18n.zh
    zh.setdefault("User Login", "用户登录")
    t = partial(wdu.tt, lang="zh", po=dict(zh=zh))
    app = wx.App()
    LoginWindow(last_user="domain\\test", password="password", t=t)
    app.MainLoop()


if __name__ == "__main__":
    test_run()
    test_i18n_run()



"""
wxbreads
----------------

Small handy snippets of wxpython
"""
import os.path

from setuptools import setup

folder = os.path.dirname(os.path.abspath(__file__))
version = author = ""
with open(os.path.join(folder, "wxbreads/__init__.py")) as f:
    for line in f:
        if line.startswith("__version__ = "):
            version = line.split("=")[-1].strip().replace('"', "")
        elif line.startswith("__author__ = "):
            author = line.split("=")[-1].strip().replace('"', "")
            break

setup(
    name="wxbreads",
    version=version.replace("'", ""),
    url="https://github.com/lixxu/wxbreads",
    license="BSD",
    author=author.replace("'", ""),
    author_email="xuzenglin@gmail.com",
    description="Small handy snippets of wxpython",
    long_description=__doc__,
    packages=["wxbreads"],
    zip_safe=False,
    platforms="any",
    install_requires=["six"],
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os.path
import tempfile
import traceback
from datetime import datetime
from functools import partial

import six
import windbreads.utils as wdu
import wx
import wx.lib.delayedresult as delayedresult

# import wx.lib.evtmgr as em
import wx.lib.scrolledpanel as scrolled
from wx.lib.busy import BusyInfo

import wxbreads.images as wxi
import wxbreads.utils as wxu
import wxbreads.widgets as wxw

SCROLLED_STYLE = wx.TAB_TRAVERSAL  # | wx.SUNKEN_BORDER
CP_STYLE = wx.CP_DEFAULT_STYLE | wx.CP_NO_TLW_RESIZE


class BaseBase(object):
    app_name = "App"
    app_title = ""
    app_size = (-1, -1)
    app_version = ""
    app_copyright = ""
    update_font = False
    quit_confirm = True
    quit_password = ""
    root_pass = "guess"
    min_chinese_fonts = 5
    remember_window = False
    show_version_in_title = True
    clear_echo_row = 0

    def init_values(self, **kwargs):
        self.opened_dlg = None
        self.need_reload = False
        if not hasattr(self, "t"):
            self.t = kwargs.get("t")

        self.destroy = kwargs.get("destroy", True)
        self.has_tray = False
        fonts = wxu.get_chinese_fonts()
        self.lang_wgts = []
        self.setting_wgts = []
        self.flat_menu = None
        self.support_chinese = len(fonts) >= self.min_chinese_fonts
        self.echo_lines = []
        self.is_echoing = False
        self.echoed_row = 0  # lines that echoed

    @property
    def screen_size(self):
        return wx.GetDisplaySize()

    def refresh_title(self, **kwargs):
        self.SetTitle(self.get_title(**kwargs))

    def highlight(self, wgt, color=None, focus=True, **kwargs):
        if color is None:
            color = wxw.HIGHLIGHT_RED

        wgt.SetBackgroundColour(color)
        if focus:
            wxw.focus_on(wgt, **kwargs)

    def get_title(self, **kwargs):
        title = kwargs.get("title") or self.app_title or self.app_name
        title = self.tt(title)
        if not self.show_version_in_title:
            return title

        version = kwargs.get("version") or self.app_version
        return "{}{}".format(title, " - " + version if version else "")

    def add_lang_wgt(self, items):
        if self.support_chinese:
            self.lang_wgts.append(items)

    def get_font(self, bold=False, size=None):
        font = self.GetFont()
        if bold:
            font.SetWeight(wx.BOLD)

        if size:
            font.SetPointSize(size)

        return font

    def get_lang_text(self, lang="en"):
        return "English" if lang == "en" else "中文"

    def auto_font(self, lang=None):
        if not self.update_font:
            return None

        en_font, ch_font = self.get_fonts()
        name = en_font if (lang or self.get_lang()) == "en" else ch_font
        return self.create_font(name)

    def create_font(self, face=None):
        if face:
            font = self.GetFont()
            font.SetFaceName(face)
            return font
            # return wx.Font(-1, wx.DEFAULT, wx.NORMAL, wx.NORMAL, False, face)

        return None

    def create_book(self, parent, id=wx.ID_ANY, **kwargs):
        return wxw.add_fnb(parent, id, **kwargs)

    def create_boxsizer(self, orient="vertical"):
        return wx.BoxSizer(wx.VERTICAL if orient[0] == "v" else wx.HORIZONTAL)

    def get_fonts(self):
        return self.get_english_font(), self.get_best_chinese_font()

    def get_chinese_fonts(self):
        return wxu.get_chinese_fonts()

    def get_best_chinese_font(self, fonts=[]):
        font = wxu.get_best_chinese_font(fonts or self.get_chinese_fonts())
        return font if font else self.get_english_font()

    def get_basic_font(self):
        return self.GetFont().GetFaceName()

    def get_english_font(self):
        if not hasattr(self, "english_font_name"):
            self.english_font_name = self.get_basic_font()

        return self.english_font_name

    def other_clock_work(self):
        pass

    def can_remember_window(self):
        logic1 = self.remember_window and hasattr(self, "config")
        return logic1 and hasattr(self, "dump_config")

    def restore_position(self):
        if self.can_remember_window() and "app_w" in self.config:
            w, h = int(self.config["app_w"]), int(self.config["app_h"])
            x, y = int(self.config["app_x"]), int(self.config["app_y"])
            self.SetSize((w, h))
            self.SetPosition((x, y))

    def other_clean_work(self):
        if self.can_remember_window():
            w, h = self.GetSize()
            x, y = self.GetPosition()
            self.config.update(
                app_w=str(w), app_h=str(h), app_x=str(x), app_y=str(y)
            )
            self.dump_config()

    def get_lang(self):
        return "en"

    def get_zh_mo(self):
        return None

    def get_en_mo(self):
        return None

    def tt(self, text):
        return wdu.ttt(text, self.t)

    def set_font(self, wgt, font=None):
        wxw.set_font(wgt, font)

    def set_label(self, wgt, label="", tooltip=""):
        wgt.SetLabel(self.tt(label))
        self.set_tooltip(wgt, tooltip)

    def set_fg(self, wgt, fg=None):
        wxw.set_fg(wgt, fg)

    def set_bg(self, wgt, bg=None):
        wxw.set_bg(wgt, bg)

    def set_tooltip(self, wgt, tooltip=""):
        wxw.set_tooltip(wgt, tooltip, self.t)

    def set_min_size(self, size=None):
        self.SetMinSize(size or self.GetSize())

    def set_max_size(self, size=None):
        self.SetMaxSize(size or self.GetSize())

    def fix_size(self, size=None):
        self.set_min_size(size)
        self.set_max_size(size)

    def show(self, **kwargs):
        cop = kwargs.get("cop")
        if cop is None:
            cop = kwargs.get("center_on_parent", True)

        cos = kwargs.get("cos")
        if cos is None:
            cos = kwargs.get("center_on_screen", True)

        if cop:  # center on parent
            self.CenterOnParent()
        elif cos:
            self.CenterOnScreen()
        elif kwargs.get("center", True):
            self.Centre(wx.BOTH)

        if kwargs.get("modal", True) and hasattr(self, "ShowModal"):
            self.ShowModal()
            return

        effect = kwargs.get("effect")
        if effect:
            self.ShowWithEffect(effect, int(kwargs.get("timeout", 0)))
        else:
            self.Show()

        try:
            self.restore_position()
        except Exception:
            pass

    def on_upper_case(self, evt=None):
        obj = evt.GetEventObject()
        text = obj.GetValue().strip().upper()
        obj.ChangeValue(text)
        obj.SetInsertionPointEnd()
        evt.Skip()

    def on_hide(self, evt=None):
        self.Hide()

    def on_quit(self, evt=None):
        wxw.quick_quit(
            self,
            t=self.t,
            need_confirm=self.quit_confirm,
            need_password=self.quit_password,
        )

    def on_flat_menu(self, evt):
        btn = evt.GetEventObject()
        size = btn.GetSize()
        pos = btn.GetParent().ClientToScreen(btn.GetPosition())
        self.flat_menu.SetOwnerHeight(size.y)
        self.flat_menu.Popup(wx.Point(pos.x, pos.y), self)

    def need_adjust_opened_dlg(self):
        return self.has_tray or self.opened_dlg is not None

    def popup(self, caption, msg, icon="i", **kwargs):
        if self.need_adjust_opened_dlg():
            self.opened_dlg += 1

        if self.has_tray:
            if kwargs.pop("restore", True):
                self.tbicon.on_restore(None)

        try:
            self.Iconize(False)
            self.Raise()
        except Exception:
            pass

        kwargs.setdefault("t", self.t)
        result = wxw.popup(
            kwargs.pop("parent", self),
            caption=caption,
            msg=msg,
            icon=icon,
            **kwargs
        )
        if self.need_adjust_opened_dlg():
            self.opened_dlg -= 1

        return result

    def bind(self, evt, func, wgt, format="self"):
        if isinstance(evt, six.string_types):
            evt = getattr(wx, evt.upper())

        if not hasattr(self, func.__name__):
            func = partial(func, self)

        if format == "self":
            self.Bind(evt, func, wgt)
        else:
            wgt.Bind(evt, func)

    def show_busyinfo(self, msg):
        return BusyInfo(self.tt(msg))

    def show_busy(self, msg, parent=None):
        wx.WindowDisabler()
        busy = wx.BusyInfo(self.tt(msg), parent=parent or self)
        wx.YieldIfNeeded()
        return busy

    def hide_busy(self, busy):
        del busy

    # def register_evt(self, listener, event, source=None, win=None, id=None):
    #     em.eventManager.Register(listener, event, source, win, id)
    #
    # def unregister_listener(self, listener):
    #     em.eventManager.DeregisterListener(listener)
    #
    # def unregister_win(self, win):
    #     em.eventManager.DeregisterWindow(win)

    def start_delay_work(self, c_func, w_func, **kwargs):
        delayedresult.startWorker(c_func, w_func, **kwargs)

    def create_scrolled_panel(self, parent=None, id=-1, style=None, **kwargs):
        return scrolled.ScrolledPanel(
            parent or self, id, style=style or SCROLLED_STYLE, **kwargs
        )

    def create_cp(self, parent, style=None, label="Show", bind=None, **kwargs):
        cp = wx.CollapsiblePane(
            parent, label=self.tt(label), style=style or CP_STYLE
        )
        if bind:
            self.Bind(wx.EVT_COLLAPSIBLEPANE_CHANGED, bind, cp)

        return cp

    def layout_scrolled_panel(self, panel, sizer=None):
        if sizer:
            self.base_layout(panel, sizer)

        panel.SetAutoLayout(True)
        panel.SetupScrolling()

    def base_layout(self, panel, sizer, with_layout=False):
        panel.SetSizer(sizer)
        sizer.Fit(self)
        if with_layout:
            panel.Layout()

    def echo_text(self, text="", **kwargs):
        wxu.echo_text(self.rtc, text, **self.set_echo_defaults(kwargs))

    def set_echo_defaults(self, kwargs):
        kwargs.setdefault("t", self.t)
        kwargs.setdefault("log_mode", "a" if six.PY2 else "ab")
        kwargs.setdefault("log_files", [])
        return kwargs

    def add_echo(self, text="", **kwargs):
        if self.clear_echo_row and kwargs.get("nl", True):
            self.echoed_row += 1
            kwargs.setdefault(
                "clear", self.echoed_row % self.clear_echo_row == 0
            )

        self.echo_lines.append((text, self.set_echo_defaults(kwargs)))

    def add_echo3(self, text="", **kwargs):
        if self.clear_echo_row and kwargs.get("nl", True):
            self.echoed_row += 1
            kwargs.setdefault(
                "clear", self.echoed_row % self.clear_echo_row == 0
            )

        self.echo_text(text, **kwargs)

    def on_echoing(self, evt=None):
        wxu.on_echoing(self)


class BaseDialog(wx.Dialog, BaseBase):
    app_name = "Dialog"

    def __init__(self, **kwargs):
        self.init_values(**kwargs)
        size = kwargs.get("size", self.app_size)

        kw = dict(size=size, title=self.get_title(**kwargs), pos=(-1, -1))
        style = kwargs.get("style")
        if style:
            kw.update(style=style)

        super(BaseDialog, self).__init__(kwargs.get("parent"), **kw)
        self.english_font_name = self.get_english_font()
        wxw.set_font(self, kwargs.get("font"))
        self.Bind(wx.EVT_CLOSE, self.on_quit)

    def setup_ui(self, parent=None, ok_text="Save", cancel_text="Cancel"):
        self.wgts = {}
        vbox = wx.BoxSizer(wx.VERTICAL)

        self.setup_other_ui(vbox)
        self.setup_ok_cancel_buttons(
            vbox, parent=parent, ok_text=ok_text, cancel_text=cancel_text
        )

        self.base_layout(self, vbox)

    def setup_ok_cancel_buttons(
        self, sizer, parent=None, ok_text="Save", cancel_text="Cancel"
    ):
        ok_btn, cancel_btn = wxw.add_ok_buttons(
            parent or self, sizer, ok_text=ok_text, cancel_text=cancel_text
        )
        ok_btn.Bind(wx.EVT_BUTTON, self.on_save)
        cancel_btn.Bind(wx.EVT_BUTTON, self.on_quit)

    def on_save(self, evt):
        pass

    def on_quit(self, evt=None):
        if self.destroy:
            self.Destroy()
        else:
            self.Hide()


class BaseWindow(wx.Frame, BaseBase):
    clock_timer_id = wx.NewIdRef()
    echo_timer_id = wx.NewIdRef()
    auth_setting = True
    app_remark = "Description for cool app"
    app_author = ""
    reset_copyright = True
    reset_copyright_seconds = (0, 1, 30, 31)
    sbar_width = [260, -1, 130]

    def __init__(self, **kwargs):
        self.init_values(**kwargs)

        self.has_sbar = False
        self.full_title = self.get_title(**kwargs)
        size = kwargs.get("size", self.app_size)
        kw = dict(size=size, title=self.full_title)
        style = kwargs.get("style")
        if style:
            kw.update(style=style)

        super(BaseWindow, self).__init__(kwargs.get("parent"), **kw)
        self.english_font_name = self.get_english_font()
        self.is_running = False
        self.logo = img = wxi.logo.GetImage()
        try:
            icon = wx.IconFromBitmap(img.ConvertToBitmap())
        except AttributeError:
            icon = wx.Icon(img.ConvertToBitmap())

        self.SetIcon(icon)
        self.Bind(wx.EVT_CLOSE, self.on_quit)

    def on_start(self, evt):
        if evt:
            evt.Skip()

    def get_changes(self):
        return {}

    def on_changes(self, evt):
        if evt:
            evt.Skip()

        changes = self.get_changes()
        if changes:
            wxw.popup_smd(self, **changes)

    def setup_base_big_buttons(self, parent, sizer, **kwargs):
        buttons = wxw.quick_big_buttons(self, parent, t=self.t, **kwargs)
        wxw.quick_pack(sizer, wgts=[(btn[0], 1) for btn in buttons])
        self.big_buttons = buttons

    def setup_base_ui(
        self, with_big=True, big_kw={}, main_kw={}, settings_kw={}
    ):
        self.panel = wx.Panel(self)
        vbox = wx.BoxSizer(wx.VERTICAL)

        if with_big:
            self.setup_base_big_buttons(self.panel, vbox, **big_kw)

        self.book = wxw.add_fnb(self.panel, wx.ID_ANY)
        self.add_base_main_page()
        self.add_base_settings_page()
        self.enable_wgts(False)

        wxw.pack(self.book, vbox, prop=1)
        self.base_layout(self.panel, vbox, with_layout=True)

    def add_base_main_page(self, title="Main", **kwargs):
        self.main_panel = p = wx.Panel(self.book)
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.fill_main_page(p, vbox, **kwargs)

        self.rtc = wxw.add_richtext(p, readonly=True, size=(-1, 450))

        wxw.pack(self.rtc, vbox, prop=1, flag="e,a")

        self.base_layout(p, vbox)
        self.book.AddPage(p, self.tt(title))

    def add_base_settings_page(self, title="Settings", **kwargs):
        self.is_locked = True
        p = wx.Panel(self.book)
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.fill_settings_page(p, vbox, **kwargs)

        line = wxw.add_line(p)
        wxw.pack(line, vbox)
        self.unlock_btn = wxw.add_button(
            p, label="Unlock", size=(-1, 40), t=self.t
        )
        cancel_btn = wxw.add_button(p, label="Cancel", size=(-1, 40), t=self.t)
        self.unlock_btn.Bind(wx.EVT_BUTTON, self.on_settings_save)
        cancel_btn.Bind(wx.EVT_BUTTON, self.on_settings_cancel)

        self.add_lang_wgt((self.unlock_btn, "Unlock"))
        self.add_lang_wgt((cancel_btn, "Cancel"))

        wxw.quick_pack(vbox, wgts=[(self.unlock_btn, 1), cancel_btn])

        self.base_layout(p, vbox)
        self.book.AddPage(p, self.tt(title))

    def fill_main_page(self, parent, sizer=None, **kwargs):
        pass

    def fill_settings_page(self, parent, sizer=None, **kwargs):
        pass

    def on_settings_save(self, evt=None):
        if self.is_locked:
            if not self.prepare_setting():
                return

            self.set_label(self.unlock_btn, "Save")
            self.enable_wgts(True)
            self.is_locked = False
        else:
            self.save_settings()
            self.on_settings_cancel()

    def save_settings(self):
        pass

    def on_settings_cancel(self, evt=None):
        self.is_locked = True
        self.set_label(self.unlock_btn, "Unlock")
        self.enable_wgts(False)

    def enable_wgts(self, enable=True, wgts=[]):
        [wgt.Enable(enable) for wgt in wgts or self.setting_wgts]

    def setup_timers(self, clock_ms=1000, echo_ms=200):
        self.all_timers = []
        if clock_ms:
            self.clock_timer = wxw.add_timer(
                self, self.clock_timer_id, self.on_clock_tick, clock_ms
            )
            self.all_timers.append(self.clock_timer)

        if echo_ms:
            self.echo_timer = wxw.add_timer(
                self, self.echo_timer_id, self.on_echoing, echo_ms
            )

            self.all_timers.append(self.echo_timer)

    def stop_timers(self, delete=False):
        if hasattr(self, "all_timers"):
            wxu.stop_timers(self.all_timers, delete=delete)

    def on_clock_tick(self, evt=None):
        if self.has_sbar:
            wxu.update_clock_statusbar(self.sbar, idx=self.sb_count - 1)
            if self.reset_copyright:
                if datetime.now().second in self.reset_copyright_seconds:
                    self.update_status(self.get_copyright(), 0, t=None)

        self.other_clock_work()

    def update_run_ts(self, idx=1, as_seconds=True):
        if self.is_running:
            time_df = datetime.now() - self.start_ts
            if as_seconds:
                self.update_status("{}s".format(time_df.total_seconds()), idx)
            else:
                self.update_status("{}".format(time_df), idx)

    def setup_statusbar(self):
        self.has_sbar = True
        self.sb_count = len(self.get_sb_width())
        self.sbar = wxw.add_statusbar(
            self, widths=self.get_sb_width(), values=self.get_sb_value()
        )
        self.SetStatusBar(self.sbar)

    def get_sb_width(self):
        """Width items for status bar."""
        return self.sbar_width

    def get_sb_value(self):
        """Value items for status bar."""
        return [self.get_copyright(), "", ""]

    def get_copyright(self):
        return self.app_copyright or wdu.get_copy_right()

    def update_status(self, text, idx, **kwargs):
        kwargs.setdefault("t", self.t)
        try:
            wxu.set_status_text(self.sbar, text, idx, **kwargs)
        except Exception:
            temp_file = os.path.join(tempfile.gettempdir(), "wxbugs.txt")
            with open(temp_file, "a") as f:
                f.write("{}\n".format(traceback.format_exc()))

    def setup_tray(self, **kwargs):
        import wxbreads.trayicon as wxt

        try:
            args = dict(
                icon=kwargs.get("icon") or self.logo,
                text=kwargs.get("text", self.app_title or self.app_name),
                t=self.t,
            )
            self.tbicon = wxt.TrayIcon(self, **args)
            self.opened_dlg = 0
            self.has_tray = True
        except Exception:
            self.tbicon = None

    def update_t(self):
        wdu.update_t(
            self,
            lang=self.get_lang(),
            zh=self.get_zh_mo(),
            en=self.get_en_mo(),
        )

    def refresh_translation(self, wgts=[]):
        font = self.auto_font()
        for lwgt in wgts:
            tooltip = ""
            if len(lwgt) == 2:
                wgt, label = lwgt
            else:
                wgt, label, tooltip = lwgt

            wxw.set_tooltip(wgt, tooltip, self.t)
            if label is not None:
                self.set_label(wgt, label)

            wxw.set_font(wgt, font)

        if hasattr(self, "panel"):
            self.panel.Layout()
        else:
            self.Layout()

        self.Refresh()

    def on_setting(self, evt=None):
        if self.prepare_setting():
            self.open_setting()

    def open_setting(self):
        raise NotImplementedError

    def prepare_setting(self):
        if self.is_running:
            self.popup("Warning", "Please stop current running task", "w")
            return False

        if hasattr(self, "is_root") and self.is_root:
            return True

        if not self.auth_setting:
            return True

        if self.need_adjust_opened_dlg():
            self.opened_dlg += 1

        if hasattr(self, "own_security_check"):
            check_ok = self.own_security_check()
        else:
            check_ok = wxw.quick_password_entry(
                self, root_pass=self.root_pass, t=self.t
            )

        if self.need_adjust_opened_dlg():
            self.opened_dlg -= 1

        return check_ok

    def on_about(self, evt=None):
        if self.need_adjust_opened_dlg():
            self.opened_dlg += 1

        wxw.quick_about(**self.get_about_kwargs())
        if self.need_adjust_opened_dlg():
            self.opened_dlg -= 1

    def get_about_icon(self):
        try:
            return wxi.phoenix.GetIcon()
        except Exception:
            return wxi.phoenix.getIcon()

    def get_about_extra_kwargs(self):
        return {}

    def get_about_kwargs(self):
        kw = dict(
            name=self.app_title or self.app_name,
            version=self.app_version,
            icon=self.get_about_icon(),
            remark=self.app_remark,
            t=self.t,
            website=getattr(self, "app_website", ""),
            author=getattr(self, "app_author", ""),
            licence=getattr(self, "app_licence", ""),
            copyright=self.get_copyright(),
        )
        kw.update(self.get_about_extra_kwargs())
        return kw

    def need_adjust_opened_dlg(self):
        return self.has_tray or self.opened_dlg is not None

    def on_popup_lang(self, evt):
        wxu.on_popup_lang(self, evt)

    def save_lang(self, lang=None):
        raise NotImplementedError

    def popup_lang(self, evt):
        wxu.popup_lang(self, evt)

    def update_ui_lang(self, refresh=True):
        wxu.update_ui_lang(self, refresh)


BaseFrame = BaseWindow


def run_app(window_class, *args, **kwargs):
    app = wx.App()
    window_class(*args, **kwargs)
    app.MainLoop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import sys
import threading

import six
import wx
import wx.lib.agw.flatmenu as FM
import wx.lib.agw.flatnotebook as fnb
import wx.lib.agw.genericmessagedialog as gmd
import wx.lib.dialogs
import wx.lib.masked as masked
import wx.richtext as rt

OLD_WX = True
try:
    AboutDialogInfo = wx.AboutDialogInfo
    AboutBox = wx.AboutBox
    DatePickerCtrl = wx.DatePickerCtrl
except AttributeError:
    import wx.adv

    AboutDialogInfo = wx.adv.AboutDialogInfo
    AboutBox = wx.adv.AboutBox
    DatePickerCtrl = wx.adv.DatePickerCtrl

    OLD_WX = False

import windbreads.utils as wdu

import wxbreads.utils as wxu

"""Remark:
fsize/fstyle/fkw: 1st widget size/style/kwargs
ssize/sstyle/skw: 2nd widget size/sstyle/kwargs
tsize/tstyle/tkw: 3rd widget size/tstyle/kwargs
"""

ICONS = dict(
    info=wx.ICON_INFORMATION,
    i=wx.ICON_INFORMATION,
    warn=wx.ICON_WARNING,
    warning=wx.ICON_WARNING,
    w=wx.ICON_WARNING,
    error=wx.ICON_ERROR,
    e=wx.ICON_ERROR,
    question=wx.ICON_QUESTION,
    q=wx.ICON_QUESTION,
    exclamation=wx.ICON_EXCLAMATION,
    ex=wx.ICON_EXCLAMATION,
)
DEFAULT_WILDCARD = "All files|*"
HIGHLIGHT_RED = "#F75D59"
if OLD_WX:
    ABOUT_FORMAT = "{}\n\nPlatform:\n- Python {}\n- wxPython {} ({})\n- {}\n"
else:
    ABOUT_FORMAT = "{}\n\nPlatform:\n- Python {}\n- wxPython {}\n  * ({})\
    \n- {}\n"

FLAT_MENU_EVENTS = dict(
    selected=FM.EVT_FLAT_MENU_SELECTED,
    dismissed=FM.EVT_FLAT_MENU_DISMISSED,
    mouse_out=FM.EVT_FLAT_MENU_ITEM_MOUSE_OUT,
    mouse_over=FM.EVT_FLAT_MENU_ITEM_MOUSE_OVER,
)


def create_id():
    return wx.NewIdRef()


def set_tooltip(wgt, tooltip="", t=None):
    if tooltip:
        func = wgt.SetToolTipString if OLD_WX else wgt.SetToolTip
        func(wdu.ttt(tooltip, t=t))


def set_fg(wgt, fg=None):
    if fg:
        wgt.SetForegroundColour(fg)


def set_bg(wgt, bg=None):
    if bg:
        wgt.SetBackgroundColour(bg)


def set_font(wgt, font=None):
    if font:
        wgt.SetFont(font)


def focus_on(wgt, select=False, clear=False):
    if clear:
        wgt.Clear()
    elif select:
        wgt.SetSelection(-1, -1)

    wgt.SetFocus()


def cat_msg_with_args(msg, **kwargs):
    kargs = kwargs.pop("kargs", {})
    if kargs:
        msg = msg.format(**kargs)

    args = kwargs.pop("args", None)
    if args:
        if not isinstance(args, (tuple, list)):
            args = (args,)

        msg = msg.format(*args)

    return msg


def popup(parent=None, caption="caption", **kwargs):
    t = kwargs.get("t")
    font = wxu.auto_get_font(parent, **kwargs)
    btn = kwargs.pop("btn", wx.OK)
    need_return = kwargs.pop("need_return", False)
    size = kwargs.pop("size", (-1, -1))
    icon = kwargs.pop("icon", "i")
    msg = kwargs.pop("msg", "")
    icon = ICONS.get(icon, ICONS["i"])
    if isinstance(msg, six.string_types):
        if not isinstance(msg, six.text_type):
            umsg = msg.decode(wdu.detect_encoding(msg)["encoding"])
        else:
            umsg = msg

    else:
        umsg = "{}".format(msg)

    if t:
        umsg = wdu.ttt(umsg, t)
        title = wdu.ttt(caption, t)
    else:
        title = caption

    umsg = cat_msg_with_args(umsg, **kwargs)
    dlg_kw = dict()
    if OLD_WX or threading.currentThread().name == "MainThread":
        dlg_cls = gmd.GenericMessageDialog
        dlg_kw.update(size=size)
    else:
        dlg_cls = wx.MessageDialog

    dlg = dlg_cls(parent, umsg, title, btn | icon, **dlg_kw)
    help_label = kwargs.get("help_label", "Help")
    ok_label = kwargs.get("ok_label", "OK")
    cancel_label = kwargs.get("cancel_label", "Cancel")
    yes_label = kwargs.get("yes_label", "Yes")
    no_label = kwargs.get("no_label", "No")
    if t:
        help_label = wdu.ttt(help_label, t)
        ok_label = wdu.ttt(ok_label, t)
        cancel_label = wdu.ttt(cancel_label, t)
        yes_label = wdu.ttt(yes_label, t)
        no_label = wdu.ttt(no_label, t)

    dlg.SetHelpLabel(help_label)
    dlg.SetOKLabel(ok_label)
    dlg.SetOKCancelLabels(ok_label, cancel_label)
    dlg.SetYesNoLabels(yes_label, no_label)
    dlg.SetYesNoCancelLabels(yes_label, no_label, cancel_label)
    dlg.SetMessage(umsg)
    set_font(dlg, font)
    if need_return:
        return dlg

    dlg.CenterOnParent()
    opened_dlg = False
    if parent and hasattr(parent, "opened_dlg"):
        if parent.opened_dlg is not None:
            opened_dlg = True

    if opened_dlg:
        parent.opened_dlg += 1

    result = dlg.ShowModal()
    dlg.Destroy()
    if opened_dlg:
        parent.opened_dlg -= 1

    return result


def popup_smd(parent=None, msg="", caption="Message", **kwargs):
    t = kwargs.get("t")
    btn_label = kwargs.get("btn_label", "OK")
    if isinstance(msg, six.string_types):
        if not isinstance(msg, six.text_type):
            umsg = msg.decode(wdu.detect_encoding(msg)["encoding"])
        else:
            umsg = msg

    else:
        umsg = "{}".format(msg)

    if t:
        btn_label = wdu.ttt(btn_label, t)
        umsg = wdu.ttt(umsg, t)
        title = wdu.ttt(caption, t)
    else:
        title = caption

    umsg = cat_msg_with_args(umsg, **kwargs)
    font = wxu.auto_get_font(parent, **kwargs)
    dlg = wx.lib.dialogs.ScrolledMessageDialog(parent, umsg, title)
    try:
        dlg.GetChildren()[1].SetLabel(btn_label)
        set_font()
    except Exception:
        pass

    if parent and hasattr(parent, "opened_dlg"):
        if parent.opened_dlg is not None:
            parent.opened_dlg += 1

    set_font(dlg, font)
    dlg.ShowModal()
    if parent and hasattr(parent, "opened_dlg"):
        if parent.opened_dlg is not None:
            parent.opened_dlg -= 1

    dlg.Destroy()


def add_button(parent, id=-1, **kwargs):
    t = kwargs.pop("t", None)
    label = kwargs.pop("label", "Button")
    size = kwargs.pop("size", (-1, -1))
    tooltip = kwargs.pop("tooltip", "")
    font = kwargs.pop("font", None)
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    icon = kwargs.pop("icon", None)
    style = kwargs.pop("style", wx.NO_BORDER)
    kw = dict(size=size, name=kwargs.get("name", "wxButton"))
    if style:
        kw.update(style=style)

    btn = wx.Button(parent, id, wdu.ttt(label, t), **kw)
    set_tooltip(btn, tooltip, t=t)
    set_font(btn, font)
    set_fg(btn, fg)
    set_bg(btn, bg)
    if icon:
        btn.SetBitmap(icon, kwargs.get("side", wx.LEFT))
        btn.SetBitmapMargins(kwargs.get("margins", (2, 2)))

    return btn


def add_label(parent, id=-1, sizer=None, **kwargs):
    t = kwargs.get("t")
    label = kwargs.pop("label", "")
    font = kwargs.pop("font", None)
    size = kwargs.pop("size", (-1, -1))
    tooltip = kwargs.pop("tooltip", "")
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    style = kwargs.pop("style", 0)
    nargs = dict(size=size, name=kwargs.get("name", "wxStaticText"))
    if style:
        nargs.update(style=style)

    lbl = wx.StaticText(parent, id, wdu.ttt(label, t), **nargs)
    set_tooltip(lbl, tooltip, t)
    set_font(lbl, font)
    set_fg(lbl, fg)
    set_bg(lbl, bg)
    if sizer:
        pack(lbl, sizer, **kwargs.get("pack_kw", {}))

    return lbl


def add_textctrl(parent, id=-1, **kwargs):
    t = kwargs.pop("t", None)
    value = kwargs.pop("value", "")
    size = kwargs.pop("size", (-1, -1))
    style = kwargs.pop("style", 0)
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    font = kwargs.pop("font", None)
    tooltip = kwargs.pop("tooltip", "")
    multiline = kwargs.pop("multiline", False)

    nargs = dict(size=size, name=kwargs.get("name", "wxTextCtrl"))
    sty = wx.TE_MULTILINE if multiline else None
    if style and sty is not None:
        nargs.update(style=style | sty)
    elif style:
        nargs.update(style=style)
    elif sty is not None:
        nargs.update(style=sty)

    wgt = wx.TextCtrl(parent, id, "{}".format(value), **nargs)
    set_tooltip(wgt, tooltip, t)
    set_font(wgt, font)
    set_fg(wgt, fg)
    set_bg(wgt, bg)
    return wgt


def add_spinctrl(parent, id=-1, **kwargs):
    value = kwargs.pop("value", "")
    size = kwargs.pop("size", (-1, -1))
    min_value = kwargs.pop("min_value", 0)
    max_value = kwargs.pop("max_value", 100)
    init_value = kwargs.pop("init_value", 0)
    tooltip = kwargs.pop("tooltip", "")
    t = kwargs.pop("t", None)

    sc = wx.SpinCtrl(
        parent, id, value, size=size, name=kwargs.pop("name", "wxSpinCtrl")
    )
    sc.SetRange(min_value, max_value)
    sc.SetValue(init_value)
    set_tooltip(sc, tooltip, t)
    return sc


def add_spinctrl_double(parent, id=-1, **kwargs):
    value = kwargs.pop("value", "")
    size = kwargs.pop("size", (-1, -1))
    inc = kwargs.pop("inc", 1)
    min_value = kwargs.pop("min_value", 0.0)
    max_value = kwargs.pop("max_value", 100.0)
    init_value = kwargs.pop("init_value", 0)
    sc = wx.SpinCtrlDouble(
        parent,
        id,
        value,
        size=size,
        inc=inc,
        name=kwargs.pop("name", "wxSpinCtrlDouble"),
    )
    sc.SetRange(min_value, max_value)
    sc.SetValue(init_value)
    return sc


def add_richtext(parent, id=-1, **kwargs):
    value = kwargs.pop("value", "")
    size = kwargs.pop("size", (-1, -1))
    readonly = kwargs.pop("readonly", True)
    style = kwargs.pop("style", wx.TE_MULTILINE)
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    font = kwargs.pop("font", None)
    if readonly:
        style |= wx.TE_READONLY

    kw = dict(size=size, name=kwargs.pop("name", "wxRichText"), style=style)
    rtc = rt.RichTextCtrl(parent, id, value, **kw)
    set_font(rtc, font)
    set_fg(rtc, fg)
    set_bg(rtc, bg)
    return rtc


def quick_add_rtc(
    parent, sizer, size=(-1, 200), readonly=True, prop=1, flag="e,a"
):
    rtc = add_richtext(parent, size=size, readonly=readonly)
    pack(rtc, sizer, prop=prop, flag=flag)
    return rtc


def add_checkbox(parent, id=-1, **kwargs):
    t = kwargs.pop("t", None)
    label = kwargs.pop("label", "")
    size = kwargs.pop("size", (-1, -1))
    value = kwargs.pop("value", True)
    font = kwargs.pop("font", None)
    tooltip = kwargs.pop("tooltip", "")
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    wgt = wx.CheckBox(
        parent,
        id,
        wdu.ttt(label, t),
        size=size,
        name=kwargs.pop("name", "wxCheckBox"),
        style=kwargs.pop("style", 0),
    )
    set_tooltip(wgt, tooltip, t)
    set_font(wgt, font)
    set_fg(wgt, fg)
    set_bg(wgt, bg)
    wgt.SetValue(value)
    return wgt


def add_path_picker(parent, id=-1, msg="Select a directory", **kwargs):
    t = kwargs.pop("t", None)
    kind = kwargs.pop("kind", "dir")
    tooltip = kwargs.pop("tooltip", "")
    multiline = kwargs.pop("multiline", False)
    btn_label = kwargs.pop("btn_label", "")
    size = kwargs.pop("size", (-1, -1))
    prop = kwargs.pop("prop", 2)
    use_tc = kwargs.pop("use_tc", True)
    value = kwargs.pop("value", "")
    text_editable = kwargs.pop("text_editable", False)
    btn_enable = kwargs.pop("btn_enable", True)
    tc_bg = kwargs.pop("tc_bg", "white")
    tc_name = kwargs.pop("tc_name", "wxTextCtrl")
    btn_name = kwargs.pop("btn_name", "wxButton")
    wgt_cls = wx.DirPickerCtrl if kind == "dir" else wx.FilePickerCtrl
    if use_tc:
        kwargs.update(style=wx.DIRP_USE_TEXTCTRL)

    pc = wgt_cls(
        parent,
        id,
        message=wdu.ttt(msg, t),
        size=size,
        path=value,
        name=kwargs.pop("name", "wxDirPickerCtrl"),
        **kwargs
    )
    if pc.HasTextCtrl():
        pc.SetTextCtrlProportion(prop)

    tc, btn = pc.Sizer.GetChildren()
    tc = tc.GetWindow()
    btn = btn.GetWindow()
    if multiline:
        tc.SetWindowStyle(wx.TE_MULTILINE)

    btn.SetLabel(wdu.ttt(btn_label or btn.GetLabel(), t))
    if tooltip:
        tooltip = wdu.ttt(tooltip, t)
        set_tooltip(tc, tooltip)
        set_tooltip(btn, tooltip)

    tc.SetEditable(text_editable)
    set_bg(tc, tc_bg)

    btn.Enable(btn_enable)
    tc.SetName(tc_name)
    btn.SetName(btn_name)
    tc.SetInsertionPoint(0)
    return pc, tc, btn


def add_dir_picker(parent, id=-1, msg="Select a directory", **kwargs):
    kwargs.setdefault("name", "wxDirPickerCtrl")
    return add_path_picker(parent, id, kind="dir", msg=msg, **kwargs)


def add_file_picker(parent, id=-1, msg="Select a file", **kwargs):
    kwargs.setdefault("name", "wxFilePickerCtrl")
    return add_path_picker(parent, id, kind="file", msg=msg, **kwargs)


def select_open_dir(parent, title="Select a directory", **kwargs):
    t = kwargs.pop("t", None)
    style = kwargs.pop("style", wx.DD_DEFAULT_STYLE)
    dlg = wx.DirDialog(parent, wdu.ttt(title, t), style=style, **kwargs)
    folder = dlg.GetPath() if dlg.ShowModal() == wx.ID_OK else None
    dlg.Destroy()
    return folder


def select_open_file(parent, msg="Select a file", **kwargs):
    wildcard = kwargs.pop("wildcard", DEFAULT_WILDCARD)
    style = kwargs.pop("style", wx.FD_OPEN)
    multi = kwargs.pop("multi", False)
    if multi:
        style |= wx.FD_MULTIPLE

    if kwargs.pop("exist", True):
        style |= wx.FD_FILE_MUST_EXIST

    if kwargs.pop("change_dir", False):
        style |= wx.FD_CHANGE_DIR

    t = kwargs.pop("t", None)
    dlg = wx.FileDialog(
        parent,
        message=wdu.ttt(msg, t),
        defaultDir=kwargs.pop("default_dir", ""),
        defaultFile=kwargs.pop("default_file", ""),
        wildcard=wdu.ttt(wildcard, t),
        style=style,
        name=kwargs.pop("name", "wxFileDialog"),
        **kwargs
    )

    paths = dlg.GetPaths() if dlg.ShowModal() == wx.ID_OK else []
    dlg.Destroy()
    if paths:
        return paths if multi else paths[0]

    return None


def select_save_file(parent, msg="Save file as...", **kwargs):
    wildcard = kwargs.pop("wildcard", DEFAULT_WILDCARD)
    style = kwargs.pop("style", wx.FD_SAVE)
    if kwargs.pop("overwrite_prompt", True):
        style |= wx.FD_OVERWRITE_PROMPT

    if kwargs.pop("change_dir", False):
        style |= wx.FD_CHANGE_DIR

    t = kwargs.pop("t", None)
    dlg = wx.FileDialog(
        parent,
        message=wdu.ttt(msg, t),
        defaultDir=kwargs.pop("default_dir", ""),
        defaultFile=kwargs.pop("default_file", ""),
        wildcard=wdu.ttt(wildcard, t),
        style=style,
        **kwargs
    )
    path = dlg.GetPath() if dlg.ShowModal() == wx.ID_OK else None
    dlg.Destroy()
    return path


def add_staticbox(parent, id=-1, label="", orient="v", **kwargs):
    box = wx.StaticBox(
        parent,
        id,
        wdu.ttt(label, kwargs.pop("t", None)),
        name=kwargs.pop("name", "wxStaticBox"),
    )
    top_bd, other_bd = box.GetBordersForSizer()
    sizer = wx.BoxSizer(wx.VERTICAL if orient == "v" else wx.HORIZONTAL)
    sizer.AddSpacer(top_bd)
    # box.SetSizer(sizer)
    # style = wx.VERTICAL if orient == "v" else wx.HORIZONTAL
    # sizer = wx.StaticBoxSizer(box, style)
    return box, sizer


def update_widget_kwargs(kw, **kwargs):
    [kw.setdefault(k, v) for k, v in kwargs.items()]


def add_text_row(parent, sizer=None, **kwargs):
    t = kwargs.pop("t", None)
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    font = kwargs.pop("font", None)
    tooltip = kwargs.pop("tooltip", "")

    nargs = dict(tooltip=tooltip, font=font, fg=fg, bg=bg, t=t)

    # new format kwargs
    kwargs.pop("ttw", None)  # not used here
    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        size=kwargs.pop("fsize", (-1, -1)),
        style=kwargs.pop("fstyle", 0),
        name=kwargs.pop("fname", "wxStaticText"),
        label=kwargs.pop("label", ""),
        **nargs
    )

    skw = kwargs.pop("skw", {})
    update_widget_kwargs(
        skw,
        size=kwargs.pop("ssize", (-1, -1)),
        style=kwargs.pop("sstyle", 0),
        name=kwargs.pop("sname", "wxTextCtrl"),
        value=kwargs.pop("value", ""),
        multiline=kwargs.pop("multiline", False),
        **nargs
    )

    lbl = add_label(parent, **fkw)
    wgt = add_textctrl(parent, **skw)
    quick_pack(sizer, wgts=[lbl, wgt], **kwargs)

    return lbl, wgt


def add_checkbox_row(parent, sizer=None, **kwargs):
    t = kwargs.pop("t", None)
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    font = kwargs.pop("font", None)
    tooltip = kwargs.pop("tooltip", "")
    nargs = dict(tooltip=tooltip, font=font, fg=fg, bg=bg, t=t)

    # new format kwargs
    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        size=kwargs.pop("fsize", (-1, -1)),
        style=kwargs.pop("fstyle", 0),
        name=kwargs.pop("fname", "wxStaticText"),
        label=kwargs.pop("label", ""),
        **nargs
    )
    skw = kwargs.pop("skw", {})
    update_widget_kwargs(
        skw,
        size=kwargs.pop("ssize", (-1, -1)),
        style=kwargs.pop("sstyle", 0),
        name=kwargs.pop("sname", "wxCheckBox"),
        value=kwargs.pop("value", True),
        label=kwargs.pop("cb_label", ""),
        **nargs
    )
    kwargs.pop("tkw", None)  # not used here

    lbl = add_label(parent, **fkw)
    wgt = add_checkbox(parent, **skw)
    quick_pack(sizer, wgts=[lbl, wgt], **kwargs)

    return lbl, wgt


def add_combobox(parent, sizer=None, **kwargs):
    t = kwargs.pop("t", None)
    label = kwargs.pop("label", "")
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    font = kwargs.pop("font", None)
    tooltip = kwargs.pop("tooltip", "")
    readonly = kwargs.pop("readonly", False)

    kwargs.pop("tkw", None)
    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        size=kwargs.pop("fsize", (-1, -1)),
        name=kwargs.pop("fname", "wxStaticText"),
        style=kwargs.pop("fstyle", 0),
        t=t,
        label=label,
        font=font,
        fg=fg,
        bg=bg,
        tooltip=tooltip,
    )

    skw = kwargs.pop("skw", {})
    sstyle = kwargs.pop("sstyle", wx.CB_DROPDOWN | wx.CB_SORT)
    if readonly and sstyle:
        sstyle = sstyle | wx.CB_READONLY

    update_widget_kwargs(
        skw,
        size=kwargs.pop("ssize", (-1, -1)),
        value="{}".format(kwargs.pop("value", "")),
        name=kwargs.pop("sname", "wxComboBox"),
        style=sstyle,
        choices=kwargs.pop("choices", []),
    )

    lbl = None
    if label is not None:
        lbl = add_label(parent, **fkw)

    wgt = wx.ComboBox(parent, -1, **skw)
    set_tooltip(wgt, tooltip, t=t)
    quick_pack(sizer, wgts=[lbl, wgt], **kwargs)
    return lbl, wgt


def add_choice(parent, sizer=None, **kwargs):
    t = kwargs.pop("t", None)
    label = kwargs.pop("label", "")
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    font = kwargs.pop("font", None)
    value = kwargs.pop("value", "")
    tooltip = kwargs.pop("tooltip", "")

    kwargs.pop("tkw", None)
    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        label=label,
        size=kwargs.pop("fsize", (-1, -1)),
        style=kwargs.pop("fstyle", 0),
        name=kwargs.pop("fname", "wxStaticText"),
        fg=fg,
        bg=bg,
        font=font,
        tooltip=tooltip,
        t=t,
    )

    choices = kwargs.pop("choices", [])
    skw = kwargs.pop("skw", {})
    update_widget_kwargs(
        skw,
        choices=choices,
        size=kwargs.pop("ssize", (-1, -1)),
        name=kwargs.pop("sname", "wxChoice"),
    )

    lbl = None
    if label is not None:
        lbl = add_label(parent, **fkw)

    wgt = wx.Choice(parent, -1, **skw)
    wgt.SetSelection(choices.index(value) if value in choices else 0)
    set_tooltip(wgt, tooltip, t)
    wgts = [lbl, wgt] if lbl else [wgt]
    quick_pack(sizer, wgts=wgts, **kwargs)

    return lbl, wgt


def add_radio_box(parent, sizer=None, **kwargs):
    t = kwargs.pop("t", None)
    label = kwargs.pop("label", "")
    size = kwargs.pop("size", (-1, -1))
    style = kwargs.pop("style", wx.RA_SPECIFY_COLS)
    choices = kwargs.pop("choices", [])
    cols = kwargs.pop("cols", 0)
    tooltip = kwargs.pop("tooltip", "")
    value = kwargs.pop("value", 0)
    wgt = wx.RadioBox(
        parent,
        -1,
        wdu.ttt(label, t),
        size=size,
        choices=choices,
        majorDimension=cols,
        style=style,
        name=kwargs.pop("name", "wxRadioBox"),
    )
    wgt.SetSelection(int(value))
    set_tooltip(wgt, tooltip, t)
    pack(wgt, sizer, **kwargs)
    return wgt


def add_datepicker(parent, sizer=None, **kwargs):
    t = kwargs.pop("t", None)

    tooltip = kwargs.pop("tooltip", "")
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    font = kwargs.pop("font", None)
    nargs = dict(tooltip=tooltip, font=font, fg=fg, bg=bg, t=t)

    kwargs.pop("tkw", None)
    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        size=kwargs.pop("fsize", (-1, -1)),
        style=kwargs.pop("fstyle", 0),
        name=kwargs.pop("fname", "wxStaticText"),
        label=kwargs.pop("label", ""),
        **nargs
    )

    sstyle = kwargs.pop("sstyle", 0)
    style = sstyle or kwargs.pop("style", 0)
    if not style:
        dropdown = kwargs.pop("dropdown", True)
        sty1 = wx.DP_DROPDOWN if dropdown else None

        show_century = kwargs.pop("show_century", True)
        sty2 = wx.DP_SHOWCENTURY if show_century else None

        if sty1 or sty2:
            if sty1:
                style = sty1
                if sty2:
                    style |= sty2

            else:
                style = sty2
        else:
            style = wx.DP_DEFAULT | wx.DP_SHOWCENTURY

    allow_none = kwargs.pop("allow_none", False)
    if allow_none:
        style |= wx.DP_ALLOWNONE

    value = kwargs.pop("value", "")
    skw = kwargs.pop("skw", {})
    update_widget_kwargs(
        skw,
        size=kwargs.pop("ssize", (-1, -1)),
        name=kwargs.pop("sname", "wxDatePickerCtrl"),
        style=style,
    )

    lbl = add_label(parent, **fkw)
    if value:
        if isinstance(value, six.string_types):
            try:
                from dateutil.parser import parse

                value = wxu.pydate2wxdate(parse(value))
                update_widget_kwargs(skw, dt=value)
            except Exception:
                pass

        else:
            update_widget_kwargs(skw, dt=value)

    wgt = DatePickerCtrl(parent, **skw)
    set_tooltip(wgt, tooltip, t)
    quick_pack(sizer, wgts=[lbl, wgt], **kwargs)

    return lbl, wgt


def add_timectrl(parent, **kwargs):
    kwargs["min"] = kwargs.pop("min_value", None)
    kwargs["max"] = kwargs.pop("max_value", None)
    kwargs["spinButton"] = kwargs.pop("spin_btn", None)
    kwargs["fmt24hr"] = kwargs.pop("format_24", True)
    wgt = masked.TimeCtrl(parent, **kwargs)
    return wgt


def add_open_dialog(parent, sizer, label="Select folder", **kwargs):
    t = kwargs.pop("t", None)
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    font = kwargs.pop("font", None)
    tooltip = kwargs.pop("tooltip", "")

    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        size=kwargs.pop("fsize", (-1, -1)),
        label=label,
        t=t,
        fg=fg,
        bg=bg,
        font=font,
        tooltip=tooltip,
    )

    skw = kwargs.pop("skw", {})
    update_widget_kwargs(
        skw,
        size=kwargs.pop("ssize", (-1, -1)),
        value=kwargs.pop("value", ""),
        multiline=kwargs.pop("multiline", False),
        t=t,
        tooltip=tooltip,
        fg=fg,
        bg=bg,
        font=font,
    )

    tkw = kwargs.pop("tkw", {})
    update_widget_kwargs(
        tkw,
        label=kwargs.pop("btn_label", "Browse"),
        id=kwargs.pop("btn_id", None),
        size=kwargs.pop("tsize", (-1, -1)),
        t=t,
        tooltip=tooltip,
        fg=fg,
        bg=bg,
        font=font,
    )

    lbl = add_label(parent, **fkw)
    txt = add_textctrl(parent, **skw)
    btn = add_button(parent, **tkw)

    quick_pack(sizer, wgts=[(lbl, 0), (txt, 1), (btn, 0)], **kwargs)
    return lbl, txt, btn


def add_line(parent, id=-1, size=(-1, -1), orient="h", sizer=None, **pack_kw):
    style = wx.LI_HORIZONTAL if orient == "h" else wx.LI_VERTICAL
    line = wx.StaticLine(parent, id, size=size, style=style)
    if sizer:
        pack(line, sizer, **pack_kw)

    return line


def add_ok_buttons(parent, sizer, id=-1, size=(100, 40), border=5, **kwargs):
    t = kwargs.pop("t", None)
    pack_line = kwargs.pop("pack_line", True)
    if pack_line:
        sl = add_line(parent, id)

    kwargs.pop("tkw", None)
    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        label=kwargs.pop("ok_text", "OK"),
        size=size,
        t=t,
        name=kwargs.pop("fname", "wxButton"),
    )

    skw = kwargs.pop("skw", {})
    update_widget_kwargs(
        skw,
        label=kwargs.pop("cancel_text", "Cancel"),
        size=size,
        t=t,
        name=kwargs.pop("sname", "wxButton"),
    )

    ok_btn = add_button(parent, wx.ID_OK, **fkw)
    ok_btn.SetDefault()
    cancel_btn = add_button(parent, wx.ID_CANCEL, **skw)

    btn_sizer = wx.StdDialogButtonSizer()
    btn_sizer.AddButton(ok_btn)
    btn_sizer.AddButton(cancel_btn)
    btn_sizer.Realize()

    if pack_line:
        pack(sl, sizer, flag="e,t", border=15)

    pack(btn_sizer, sizer, prop=0, flag="ac,a", border=border)
    return ok_btn, cancel_btn


def add_statusbar(
    parent, widths=[260, -1, 130], values=["", "", ""], **kwargs
):
    t = kwargs.pop("t", None)
    sbar = wx.StatusBar(parent)
    sbar.SetFieldsCount(len(widths), widths)
    for i, v in enumerate(values):
        if v is None:
            continue

        sbar.SetStatusText(wdu.ttt(v, t), i)

    return sbar


def add_timer(self, timer_id, timer_func, miliseconds=-1, one_shot=False):
    timer = wx.Timer(self, timer_id)
    self.Bind(wx.EVT_TIMER, timer_func, id=timer_id)
    if miliseconds > 0:
        wxu.start_timer(timer, miliseconds, one_shot)

    if hasattr(self, "all_timers") and timer not in self.all_timers:
        self.all_timers.append(timer)

    return timer


def add_fnb(parent, id=-1, **kwargs):
    style = kwargs.pop("style", 0)
    active_color = kwargs.pop("active_color", None)
    no_drag = kwargs.pop("no_drag", True)
    no_x = kwargs.pop("no_x", True)
    no_nav = kwargs.pop("no_nav", True)
    has_bg = kwargs.pop("has_bg", True)
    tab_style = kwargs.pop("tab_style", "vc8").strip().upper()

    style |= wx.TAB_TRAVERSAL

    if no_drag:
        style |= fnb.FNB_NODRAG

    if no_x:
        style |= fnb.FNB_NO_X_BUTTON

    if no_nav:
        style |= fnb.FNB_NO_NAV_BUTTONS

    if has_bg:
        style |= fnb.FNB_BACKGROUND_GRADIENT

    style |= getattr(fnb, "FNB_{}".format(tab_style))
    if kwargs.pop("bottom", False):
        style |= fnb.FNB_BOTTOM

    if active_color is None:
        # and this color
        active_color = wx.Colour(0, 128, 255, 255)

    nb = fnb.FlatNotebook(parent, id, agwStyle=style)
    nb.SetActiveTabColour(active_color)
    return nb


def get_sizer_flags(flags=""):
    flag = None
    for text in (flags or "").replace(" ", "").lower().split(","):
        if text in ("e", "exp", "expand"):
            f = wx.EXPAND
        elif text in ("l", "left"):
            f = wx.LEFT
        elif text in ("r", "right"):
            f = wx.RIGHT
        elif text in ("t", "top"):
            f = wx.TOP
        elif text in ("b", "bot", "bottom"):
            f = wx.BOTTOM
        elif text in ("a", "all"):
            f = wx.ALL
        elif text in ("sh", "shaped"):
            f = wx.SHAPED
        elif text in ("fix", "fixed"):
            f = wx.FIXED_MINSIZE
        elif text in ("hide", "hidden"):
            f = wx.RESERVE_SPACE_EVEN_IF_HIDDEN
        elif text in ("ac", "center", "centre"):
            f = wx.ALIGN_CENTER
        elif text in ("al", "a_left"):
            f = wx.ALIGN_LEFT
        elif text in ("ar", "a_right"):
            f = wx.ALIGN_RIGHT
        elif text in ("at", "a_top"):
            f = wx.ALIGN_TOP
        elif text in ("ab", "a_bot", "a_bottom"):
            f = wx.ALIGN_BOTTOM
        elif text in ("acv", "a_center_v", "a_centre_v"):
            f = wx.ALIGN_CENTER_VERTICAL
        elif text in ("ach", "a_center_h", "a_centre_h"):
            f = wx.ALIGN_CENTER_HORIZONTAL
        else:
            continue

        if flag is None:
            flag = f
        else:
            flag = flag | f

    if flag is None:
        flag = wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP

    return flag


def pack(wgt, sizer="h", **kwargs):
    if sizer == "h":
        sizer = wx.BoxSizer(wx.HORIZONTAL)
    elif sizer == "v":
        sizer = wx.BoxSizer(wx.VERTICAL)
    elif not sizer:
        return

    kargs = dict(flag=get_sizer_flags(kwargs.get("flag")))
    border = kwargs.get("border", 3)
    if border:
        kargs.update(border=border)

    sizer.Add(wgt, kwargs.get("prop", 0), **kargs)
    return sizer


def sort_wgts(wgts=[], **kwargs):
    lst = []
    last_prop = kwargs.get("last_prop", 1)
    size = len(wgts)
    for i, wgt in enumerate(wgts, 1):
        if isinstance(wgt, (tuple, list)):
            lst.append(wgt)
        else:
            lst.append((wgt, last_prop if i == size else 0))

    return lst


def about_box(**kwargs):
    t = kwargs.pop("t", None)
    info = AboutDialogInfo()
    icon = kwargs.pop("icon", None)
    icon_fmt = kwargs.pop("icon_fmt", None)
    if icon:
        if isinstance(icon, six.string_types):
            args = [icon]
            if icon_fmt:
                args.append(icon_fmt)

            info.SetIcon(wx.Icon(*args))
        else:
            info.SetIcon(icon)  # may be PyEmbeddedImage

    name = kwargs.pop("name", "app name")
    version = kwargs.pop("version", "0.1")
    description = kwargs.pop("description", "description")
    copy_right = kwargs.pop("copyright", "copyright")
    info.SetName(wdu.ttt(name, t))
    info.SetVersion(wdu.ttt(version, t))
    info.SetDescription(wdu.ttt(description, t))
    info.SetCopyright(wdu.ttt(copy_right, t))
    website = kwargs.pop("website", "")
    if website:
        info.SetWebSite(website)

    licence = kwargs.pop("licence", "")
    if licence:
        info.SetLicence(licence)

    developers = kwargs.pop("developers", [])
    doc_writers = kwargs.pop("doc_writers", [])
    artists = kwargs.pop("artists", [])
    tranlators = kwargs.pop("tranlators", [])
    [info.AddDeveloper(developer) for developer in developers]
    [info.AddDocWriter(writer) for writer in doc_writers]
    [info.AddArtist(artist) for artist in artists]
    [info.AddTranslator(tranlator) for tranlator in tranlators]
    AboutBox(info)


def adjust_opened_dlg(self, inc=1):
    if hasattr(self, "opened_dlg") and self.opened_dlg is not None:
        self.opened_dlg += inc


def quick_quit(self, **kwargs):
    """Quick handy method to ask for quit."""
    need_confirm = kwargs.pop("need_confirm", True)
    need_password = kwargs.pop("need_password", "")
    other_clean_work = kwargs.pop("other_clean_work", None)
    if hasattr(self, "is_running") and self.is_running:
        caption = kwargs.pop("running_caption", "Warning")
        msg = kwargs.pop("running_msg", "Please stop current running task")
        icon = kwargs.pop("running_icon", "w")
        popup(self, caption=caption, msg=msg, icon=icon, **kwargs)
        return

    if hasattr(self, "tray") and self.opened_dlg > 0:
        caption = kwargs.pop("opened_caption", "Warning")
        msg = kwargs.pop("opened_msg", "Please close other dialogs")
        icon = kwargs.pop("opened_icon", "w")
        popup(self, caption=caption, msg=msg, icon=icon, **kwargs)
        return

    if need_password:
        adjust_opened_dlg(self, 1)
        is_ok = quick_password_entry(self, root_pass=need_password, **kwargs)
        adjust_opened_dlg(self, -1)
        if not is_ok:
            return

    elif need_confirm:
        adjust_opened_dlg(self, 1)
        answer = popup(
            self,
            caption=kwargs.pop("ask_caption", "Confirmation"),
            msg=kwargs.pop("ask_msg", "Are you sure to quit?"),
            icon=kwargs.pop("ask_icon", "q"),
            btn=kwargs.pop("btn", wx.YES_NO | wx.NO_DEFAULT),
            **kwargs
        )
        adjust_opened_dlg(self, -1)
        if answer == wx.ID_NO:
            return

    self.Hide()
    if hasattr(self, "hm"):
        try:
            self.hm.UnhookKeyboard()
        except Exception:
            pass

        try:
            self.hm.UnhookMouse()
        except Exception:
            pass

    if hasattr(self, "stop_timers"):
        self.stop_timers(delete=True)

    if hasattr(self, "tbicon") and self.tbicon is not None:
        self.tbicon.Destroy()

    if other_clean_work:
        other_clean_work()
    elif hasattr(self, "other_clean_work"):
        self.other_clean_work()

    self.Destroy()
    return True


def quick_about(*args, **kwargs):
    fmt = kwargs.pop("fmt", ABOUT_FORMAT)
    t = kwargs.get("t", None)
    copy_right = kwargs.pop("copyright", wdu.get_copy_right())
    authors = kwargs.pop("author", None)
    writers = kwargs.pop("writers", None)
    remark = kwargs.pop("remark", "about this tool")
    description = fmt.format(
        wdu.ttt(remark, t),
        sys.version.split()[0],
        wx.VERSION_STRING,
        ", ".join(wx.PlatformInfo[1:]),
        wdu.get_platform_info(),
    )
    if authors:
        if not isinstance(authors, list):
            authors = [authors]

        kwargs.update(developers=authors)

    if writers:
        if not isinstance(writers, list):
            writers = [writers]

        kwargs.update(doc_writers=writers)

    about_info = dict(
        description=description,
        copyright=copy_right.replace("&", "&&"),
        **kwargs
    )
    about_box(**about_info)


def quick_pack(sizer=None, wgts=[], orient="h", **kwargs):
    if not sizer:
        return

    ws = sort_wgts(wgts, **kwargs)
    kargs = dict(flag=kwargs.get("flag"), border=kwargs.get("border", 3))
    box = pack(ws[0][0], sizer=orient, prop=ws[0][1], **kargs)
    for wgt, pr in ws[1:]:
        pack(wgt, box, prop=pr, **kargs)

    pack(box, sizer, **kwargs)
    return box


def quick_open_file(parent, sizer=None, label="Select File", **kwargs):
    value = kwargs.pop("value", "")
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    kwargs.pop("tkw", None)
    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        size=kwargs.pop("fsize", (-1, -1)),
        label=label,
        fg=fg,
        bg=bg,
        t=kwargs.get("t"),
        name=kwargs.pop("fname", "wxStaticText"),
    )

    skw = kwargs.pop("skw", {})
    update_widget_kwargs(skw, size=kwargs.pop("ssize", (-1, -1)), **kwargs)
    lbl = add_label(parent, **fkw)
    fp, tc, btn = add_file_picker(parent, value=value, **skw)
    quick_pack(sizer, wgts=[lbl, fp])
    return lbl, fp, tc, btn


def quick_open_folder(parent, sizer=None, label="Select Folder", **kwargs):
    fg = kwargs.pop("fg", None)
    bg = kwargs.pop("bg", None)
    kwargs.pop("tkw", None)
    fkw = kwargs.pop("fkw", {})
    update_widget_kwargs(
        fkw,
        size=kwargs.pop("fsize", (-1, -1)),
        label=label,
        fg=fg,
        bg=bg,
        t=kwargs.get("t"),
        name=kwargs.pop("fname", "wxStaticText"),
    )

    skw = kwargs.pop("skw", {})
    update_widget_kwargs(
        skw,
        size=kwargs.pop("ssize", (-1, -1)),
        value=kwargs.pop("value", ""),
        **kwargs
    )
    lbl = add_label(parent, **fkw)
    fp, tc, btn = add_dir_picker(parent, **skw)
    quick_pack(sizer, wgts=[lbl, fp])
    return lbl, fp, tc, btn


def quick_entry(parent=None, caption="", msg="Enter", password=True, **kwargs):
    entry_cls = wx.PasswordEntryDialog if password else wx.TextEntryDialog
    t = kwargs.pop("t", None)
    root_pass = kwargs.pop("root_pass", "guess")
    ok_label = kwargs.pop("ok_label", None)
    cancel_label = kwargs.pop("cancel_label", None)
    style = kwargs.pop("style", None)
    pos = kwargs.pop("pos", None)
    size = kwargs.pop("size", None)
    show_help = kwargs.pop("show_help", None)
    help_extra = kwargs.pop("help_extra", {})
    kw = {}
    if style is not None:
        kw.update(style=style)

    if pos is not None:
        kw.update(pos=pos)

    if password:
        kw.update(defaultValue=kwargs.pop("value", ""))
    else:
        kw.update(value=kwargs.pop("value", ""))

    font = wxu.auto_get_font(parent, **kwargs)
    dlg = entry_cls(parent, wdu.ttt(msg, t), wdu.ttt(caption, t), **kw)
    if font is not None:
        set_font(dlg, font)

    # update button labels for i18n
    try:
        sizers = dlg.Sizer.GetChildren()
        msg_wgt = sizers[0].Sizer.GetChildren()[0].GetWindow()
        # text_wgt = sizers[1].GetWindow()
        sizers2 = sizers[2].Sizer.GetChildren()
        # line_wgt = sizers2[0].GetWindow()
        btn_sizer = sizers2[1].Sizer
        items = btn_sizer.GetChildren()
        ok_btn, c_btn = items[1].GetWindow(), items[2].GetWindow()
        ok_btn.SetLabel(wdu.ttt(ok_label or ok_btn.GetLabel(), t))
        c_btn.SetLabel(wdu.ttt(cancel_label or c_btn.GetLabel(), t))
        if show_help:
            help_label = help_extra.get("label", wdu.ttt("Help", t))
            help_btn = add_button(
                dlg, label=help_label, size=(-1, ok_btn.GetSize().height)
            )
            pack(help_btn, btn_sizer)
            help_func = help_extra.get("event_func")
            if help_func:
                dlg.Bind(wx.EVT_BUTTON, help_func, help_btn)

        if font is not None:
            [set_font(wgt, font) for wgt in (msg_wgt, ok_btn, c_btn)]

    except Exception:
        pass

    if not size:
        size = dlg.GetClientSize()

    dlg.SetMinClientSize(size)
    dlg.SetMaxClientSize(size)
    while True:
        dlg.SetFocus()
        if dlg.ShowModal() == wx.ID_OK:
            text = dlg.GetValue()
            if password:
                if text == root_pass:
                    dlg.Destroy()
                    return True

            else:
                text = text.strip()
                if text:
                    dlg.Destroy()
                    return text

            dlg.SetValue("")
            dlg.SetFocus()
            continue

        dlg.Destroy()
        return False if password else ""


def quick_password_entry(parent=None, caption="Security Check", **kwargs):
    msg = kwargs.pop("msg", "Please enter password:")
    return quick_entry(parent, caption=caption, msg=msg, **kwargs)


def quick_text_entry(parent=None, caption="Enter Something", **kwargs):
    msg = kwargs.pop("msg", "Please enter something: ")
    return quick_entry(
        parent, caption=caption, msg=msg, password=False, **kwargs
    )


def quick_choice(parent=None, msg="Please select", **kwargs):
    t = kwargs.pop("t", None)
    caption = kwargs.pop("caption", "Please select")
    choices = kwargs.pop("choices", [])
    valid_choices = kwargs.pop("valid_choices", [])
    style = kwargs.pop("style", wx.CHOICEDLG_STYLE)
    dlg = wx.SingleChoiceDialog(
        parent, wdu.ttt(msg, t), wdu.ttt(caption, t), choices, style
    )
    font = wxu.auto_get_font(parent, **kwargs)
    set_font(dlg, font)
    ok_label = kwargs.pop("ok_label", None)
    cancel_label = kwargs.pop("cancel_label", None)
    # update button labels for i18n
    try:
        wgt_list = dlg.GetChildren()
        msg_wgt, ok_btn, c_btn = wgt_list[0], wgt_list[2], wgt_list[3]
        ok_btn.SetLabel(wdu.ttt(ok_label or ok_btn.GetLabel(), t))
        c_btn.SetLabel(wdu.ttt(cancel_label or c_btn.GetLabel(), t))
        set_font(msg_wgt, font)
        set_font(ok_btn, font)
        set_font(c_btn, font)
    except Exception:
        pass

    while 1:
        if dlg.ShowModal() == wx.ID_OK:
            selected = dlg.GetStringSelection()
            if selected in valid_choices:
                dlg.Destroy()
                return selected

            continue

        break

    dlg.Destroy()
    return None


def add_start_button(self, parent, label="Start", size=(-1, 45), **kwargs):
    font = kwargs.pop("font", None)
    if not font:
        font = self.GetFont()
        font.SetWeight(wx.BOLD)

    kw = dict(size=size, font=font, t=getattr(self, "t", None))
    kw.update(kwargs)
    btn = add_button(parent, label=label, **kw)
    btn.Bind(wx.EVT_BUTTON, self.on_start)
    btn.SetFocus()
    return btn


def add_flat_menu():
    return FM.FlatMenu()


def add_flat_menu_separator(menu):
    menu.AppendSeparator()


def bind_flat_menu(obj, handler, menu_id, name="selected"):
    obj.Bind(FLAT_MENU_EVENTS[name.lower()], handler, id=menu_id)


def add_flat_menu_item(parent, menu_id, label, add_sep=True, **kwargs):
    t = kwargs.pop("t", None)
    item = FM.FlatMenuItem(parent, menu_id, wdu.ttt(label, t), **kwargs)
    parent.AppendItem(item)
    if add_sep:
        parent.AppendSeparator()

    return item


def quick_big_buttons(
    self,
    parent,
    start=True,
    setting=True,
    hide=True,
    changes=True,
    about=True,
    return_dict=False,
    **kwargs
):
    labels = kwargs.pop("labels", {})
    sizer = kwargs.pop("sizer", None)
    font = wxu.auto_get_font(self, **kwargs)
    if not font:
        font = self.GetFont()

    font.SetWeight(wx.BOLD)
    kw = dict(size=(-1, 45), font=font)
    kw.update(kwargs)
    buttons = {} if return_dict else []
    wgts = []
    if start:
        label = labels.get("start", "Start")
        start_btn = add_button(parent, label=label, **kw)
        start_btn.Bind(wx.EVT_BUTTON, self.on_start)
        wgts.append(start_btn)
        if return_dict:
            buttons.update(start=(start_btn, label))
        else:
            buttons.append((start_btn, label))

    if setting:
        label = labels.get("settings", "Settings")
        setting_btn = add_button(parent, label=label, **kw)
        setting_btn.Bind(wx.EVT_BUTTON, self.on_setting)
        wgts.append(setting_btn)
        if return_dict:
            buttons.update(settings=(setting_btn, label))
        else:
            buttons.append((setting_btn, label))

    if hide:
        label = labels.get("hide", "Hide")
        hide_btn = add_button(parent, label=label, **kw)
        hide_btn.Bind(wx.EVT_BUTTON, self.on_hide)
        wgts.append(hide_btn)
        if return_dict:
            buttons.update(hide=(hide_btn, label))
        else:
            buttons.append((hide_btn, label))

    if changes:
        label = labels.get("changes", "Changes")
        changes_btn = add_button(parent, label=label, **kw)
        changes_btn.Bind(wx.EVT_BUTTON, self.on_changes)
        wgts.append(changes_btn)
        if return_dict:
            buttons.update(changes=(changes_btn, label))
        else:
            buttons.append((changes_btn, label))

    if about:
        label = labels.get("about", "About")
        about_btn = add_button(parent, label=label, **kw)
        about_btn.Bind(wx.EVT_BUTTON, self.on_about)
        wgts.append(about_btn)
        if return_dict:
            buttons.update(about=(about_btn, label))
        else:
            buttons.append((about_btn, label))

    if sizer:
        quick_pack(sizer, wgts=[(btn, 1) for btn in wgts])

    return buttons


add_rich_text = add_richtext
add_status_bar = add_statusbar
add_date_picker = add_datepicker
add_time_ctrl = add_timectrl
add_text_ctrl = add_textctrl
add_spin_ctrl = add_spinctrl
add_spin_ctrl_double = add_spinctrl_double
add_check_box = add_checkbox
add_static_box = add_staticbox
add_check_box_row = add_checkbox_row
add_combo_box = add_combobox
add_radiobox = add_radio_box

quick_textentry = quick_text_entry
quick_open_dir = quick_open_folder
